library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;
use work.sblock2006_package.ALL;

entity devunit is
    Port (
    	north_type  	: in std_logic_vector(TYPE_SIZE - 1 downto 0);
		south_type  	: in std_logic_vector(TYPE_SIZE - 1 downto 0);
    	east_type   	: in std_logic_vector(TYPE_SIZE - 1 downto 0);
    	west_type   	: in std_logic_vector(TYPE_SIZE - 1 downto 0);
    	center_type 	: in std_logic_vector(TYPE_SIZE - 1 downto 0);

    	north_state  	: in std_logic;
    	south_state  	: in std_logic;
    	east_state   	: in std_logic;
    	west_state   	: in std_logic;
    	center_state 	: in std_logic;

		rule 				: in std_logic_vector(RULE_SIZE-1 downto 0);    	

		hit          	: out std_logic;
    	
		-- result after applying rule
    	result_type  	: out std_logic_vector(TYPE_SIZE - 1 downto 0);
		result_state 	: out std_logic;

		rst 				: in std_logic;
    	clk 				: in std_logic
	);
end devunit;

architecture Behavioral of devunit is

  ----------------------------------------------------------------------------  
  constant DC 					: std_logic_vector(TYPE_SIZE - 1 downto 0) := "11111";
  constant CHANGE 			: std_logic := '0';
  constant GROWTH 			: std_logic := '1';
  constant COPY_FROM_NORTH : std_logic_vector(1 downto 0) := "00";
  constant COPY_FROM_SOUTH : std_logic_vector(1 downto 0) := "01";
  constant COPY_FROM_EAST  : std_logic_vector(1 downto 0) := "10";
  constant COPY_FROM_WEST  : std_logic_vector(1 downto 0) := "11";

  -----------------------------------------------------------------------------
  signal rule_valid 			: std_logic;
  signal rule_type 			: std_logic;

  signal rule_north_state_dc  : std_logic;
  signal rule_south_state_dc  : std_logic;
  signal rule_east_state_dc   : std_logic;
  signal rule_west_state_dc   : std_logic;
  signal rule_center_state_dc : std_logic;

  signal rule_north_state  : std_logic;
  signal rule_south_state  : std_logic;
  signal rule_east_state   : std_logic;
  signal rule_west_state   : std_logic;
  signal rule_center_state : std_logic;

  signal rule_north_type_dc  : std_logic;
  signal rule_south_type_dc  : std_logic;
  signal rule_east_type_dc   : std_logic;
  signal rule_west_type_dc   : std_logic;
  signal rule_center_type_dc : std_logic;

  signal rule_north_type  : std_logic_vector(TYPE_SIZE - 1 downto 0);
  signal rule_south_type  : std_logic_vector(TYPE_SIZE - 1 downto 0);
  signal rule_east_type   : std_logic_vector(TYPE_SIZE - 1 downto 0);
  signal rule_west_type   : std_logic_vector(TYPE_SIZE - 1 downto 0);
  signal rule_center_type : std_logic_vector(TYPE_SIZE - 1 downto 0);

  signal rule_dont_change_state : std_logic;
  signal rule_change_state_to   : std_logic;
  signal rule_change_type_to    : std_logic_vector(TYPE_SIZE - 1 downto 0);
  signal rule_copy_from         : std_logic_vector(1 downto 0);

  -----------------------------------------------------------------------------

  signal empty1 : std_logic;
  signal empty2 : std_logic;
  signal empty3 : std_logic;
  signal empty4 : std_logic;
  signal empty5 : std_logic;

  -----------------------------------------------------------------------------

  signal ex1_rule_valid : std_logic;

  signal ex1_typeeq1 : std_logic;
  signal ex1_typeeq2 : std_logic;
  signal ex1_typeeq3 : std_logic;
  signal ex1_typeeq4 : std_logic;
  signal ex1_typeeq5 : std_logic;

  signal ex1_stateeq1 : std_logic;
  signal ex1_stateeq2 : std_logic;
  signal ex1_stateeq3 : std_logic;
  signal ex1_stateeq4 : std_logic;
  signal ex1_stateeq5 : std_logic;
  
  signal ex1_typedc1 : std_logic;
  signal ex1_typedc2 : std_logic;
  signal ex1_typedc3 : std_logic;
  signal ex1_typedc4 : std_logic;
  signal ex1_typedc5 : std_logic;
  
  signal ex1_empty : std_logic;

  signal ex1_result_type  : std_logic_vector(TYPE_SIZE - 1 downto 0);
  signal ex1_result_state : std_logic;

begin

  -----------------------------------------------------------------------------
  -- extract individual fields from rule
  -----------------------------------------------------------------------------

  rule_valid <= rule(48);
  rule_type <= rule(47);

  -----------------------------------------------------------------------------

  rule_north_state_dc <= rule(46);
  rule_north_state <= rule(45);
  rule_north_type_dc <= rule(44);
  rule_north_type <= rule(43 downto 39);

  rule_south_state_dc <= rule(38);
  rule_south_state <= rule(37);
  rule_south_type_dc <= rule(36);
  rule_south_type <= rule(35 downto 31);

  rule_east_state_dc <= rule(30);
  rule_east_state <= rule(29);
  rule_east_type_dc <= rule(28);
  rule_east_type <= rule(27 downto 23);

  rule_west_state_dc <= rule(22);
  rule_west_state <= rule(21);
  rule_west_type_dc <= rule(20);
  rule_west_type <= rule(19 downto 15);

  rule_center_state_dc <= rule(14);
  rule_center_state <= rule(13);
  rule_center_type_dc <= rule(12);
  rule_center_type <= rule(11 downto 7);

  -----------------------------------------------------------------------------

  rule_dont_change_state <= rule(6);
  rule_change_state_to <= rule(5);
  rule_change_type_to <= rule(4 downto 0);
  rule_copy_from <= rule(1 downto 0);

  -----------------------------------------------------------------------------
  -- check for hit
  -----------------------------------------------------------------------------

  -- check for empty sblocks

  empty1 <= '1' when north_type = "00000" else '0';
  empty2 <= '1' when south_type = "00000" else '0';
  empty3 <= '1' when east_type = "00000" else '0';
  empty4 <= '1' when west_type = "00000" else '0';
  empty5 <= '1' when center_type = "00000" else '0';

  -----------------------------------------------------------------------------
  process (rst, clk)

  begin

    if rst = '0' then

      ex1_rule_valid <= '0';

      ex1_stateeq1 <= '0';
      ex1_stateeq2 <= '0';
      ex1_stateeq3 <= '0';
      ex1_stateeq4 <= '0';
      ex1_stateeq5 <= '0';

      ex1_typeeq1 <= '0';
      ex1_typeeq2 <= '0';
      ex1_typeeq3 <= '0';
      ex1_typeeq4 <= '0';
      ex1_typeeq5 <= '0';

      hit <= '0';

      ex1_typedc1 <= '0';
      ex1_typedc2 <= '0';
      ex1_typedc3 <= '0';
      ex1_typedc4 <= '0';
      ex1_typedc5 <= '0';

      ex1_empty <= '0';

    elsif rising_edge (clk) then

      ex1_rule_valid <= rule_valid;

      -- don't cares

      ex1_typedc1 <= rule_north_type_dc;
      ex1_typedc2 <= rule_south_type_dc;
      ex1_typedc3 <= rule_east_type_dc;
      ex1_typedc4 <= rule_west_type_dc;
      ex1_typedc5 <= rule_center_type_dc;

      -- compare type

      if rule_north_type = north_type then
        ex1_typeeq1 <= '1';
      else
        ex1_typeeq1 <= '0';
      end if;

      if rule_south_type = south_type then
        ex1_typeeq2 <= '1';
      else
        ex1_typeeq2 <= '0';
      end if;

      if rule_east_type = east_type then
        ex1_typeeq3 <= '1';
      else
        ex1_typeeq3 <= '0';
      end if;

      if rule_west_type = west_type then
        ex1_typeeq4 <= '1';
      else
        ex1_typeeq4 <= '0';
      end if;

      if rule_center_type = center_type then
        ex1_typeeq5 <= '1';
      else
        ex1_typeeq5 <= '0';
      end if;

      -- determine state match

      if rule_north_state = north_state or rule_north_state_dc = '1' then
        ex1_stateeq1 <= '1';
      else
        ex1_stateeq1 <= '0';
      end if;
      if rule_south_state = south_state or rule_south_state_dc = '1' then
        ex1_stateeq2 <= '1';
      else
        ex1_stateeq2 <= '0';
      end if;

      if rule_east_state = east_state or rule_east_state_dc = '1' then
        ex1_stateeq3 <= '1';
      else
        ex1_stateeq3 <= '0';
      end if;

      if rule_west_state = west_state or rule_west_state_dc = '1' then
        ex1_stateeq4 <= '1';
      else
        ex1_stateeq4 <= '0';
      end if;

      if rule_center_state = center_state or rule_center_state_dc = '1' then
        ex1_stateeq5 <= '1';
      else
        ex1_stateeq5 <= '0';
      end if;

      -- check for empty sblocks (depends on rule type)

      ex1_empty <= '1';

      if rule_type = GROWTH then
        case rule_copy_from is
          when COPY_FROM_NORTH =>
            ex1_empty <= empty1;
          when COPY_FROM_SOUTH =>
            ex1_empty <= empty2;
          when COPY_FROM_EAST =>
            ex1_empty <= empty3;
          when COPY_FROM_WEST =>
            ex1_empty <= empty4;
          when others =>
            null;
        end case;
      end if;

      if rule_type = CHANGE then
        ex1_empty <= empty5;
      end if;

      -------------------------------------------------------------------------

      hit <= ex1_rule_valid and

             (not ex1_empty) and

             (ex1_typedc1 or ex1_typeeq1) and
             (ex1_typedc2 or ex1_typeeq2) and
             (ex1_typedc3 or ex1_typeeq3) and
             (ex1_typedc4 or ex1_typeeq4) and
             (ex1_typedc5 or ex1_typeeq5) and

             ex1_stateeq1 and ex1_stateeq2 and ex1_stateeq3 and
             ex1_stateeq4 and ex1_stateeq5;

    end if;

  end process;

  -----------------------------------------------------------------------------
  -- calculate result
  -----------------------------------------------------------------------------

  process (rst, clk)
  begin

    if rst = '0' then

    elsif rising_edge (clk) then

      case rule_type is

        when CHANGE =>
          ex1_result_type <= rule_change_type_to;
          ex1_result_state <= rule_change_state_to;

        when GROWTH =>
          case rule_copy_from is
            when COPY_FROM_NORTH =>
              ex1_result_type <= north_type;
              ex1_result_state <= north_state;
            when COPY_FROM_SOUTH =>
              ex1_result_type <= south_type;
              ex1_result_state <= south_state;
            when COPY_FROM_EAST =>
              ex1_result_type <= east_type;
              ex1_result_state <= east_state;
            when COPY_FROM_WEST =>
              ex1_result_type <= west_type;
              ex1_result_state <= west_state;
            when others =>
              null;
          end case;
            
        when others =>
          null;

      end case;

      if rule_dont_change_state = '1' then
        ex1_result_state <= center_state;
      end if;

      result_type <= ex1_result_type;
      result_state <= ex1_result_state;

    end if;

  end process;
end Behavioral;
