-------------------------------------------------------------------------------
-- Title      : Sblock
-- Project    : sblock2006
-------------------------------------------------------------------------------
-- File       : sblock.vhd
-- Author     : Mats Jrgen yan  <matsjorg@orakel.ntnu.no>
-- Platform   : BenBlue-III, Virtex-II Pro (XC2VP70)
-------------------------------------------------------------------------------
-- Description: En komplett Sblock-celle
-------------------------------------------------------------------------------

library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;
use work.sblock2006_package.ALL;
library UNISIM;
use UNISIM.VComponents.all;

entity sblock is

    Port ( 
	 		  --Type-koblinger til naboene
			  Ntype 				: in std_logic_vector(4 downto 0);
           Stype 				: in std_logic_vector(4 downto 0);
           Wtype 				: in std_logic_vector(4 downto 0);
           Etype 				: in std_logic_vector(4 downto 0);
           
			  --Tilstands-kobling til naboer
			  Nval 				: in std_logic;
           Sval 				: in std_logic;
           Wval 				: in std_logic;
           Eval 				: in std_logic;    	
           
			  --Klokke og reset
			  clk 				: in std_logic;
		 	  rst 				: in std_logic;
           
			  --Kontrollsignaler for tilstand
			  dev_enable 		: in std_logic;
           Step_enable 		: in std_logic;
		     config_enable 	: in std_logic;
		     
			  --Regel fra regellager
			  rule 				: in std_logic_vector(RULE_SIZE-1 downto 0);
				
           --Konfigurasjon av Type-LUT mappingminne
		 	  t2lc_data_in		: in std_logic_vector(31 downto 0);
		     t2lc_we 			: in std_logic;		 
		     t2lc_addr_in		: in std_logic_vector(4 downto 0);

			  --Data for setting av tilstand og typen til Sblocken
		     set_enable		: in std_logic;
		     set_value			: in std_logic;
		     set_type			: in std_logic_vector(TYPE_SIZE-1 downto 0);

			  --Utgang som sier om Sblocken aktiverer en regel
		     rule_used			: out std_logic;
           
			  --Utverdi og type
			  Outval 			: out std_logic;
           Outtype 			: out std_logic_vector(4 downto 0)
		 );
end sblock;

architecture sblock_arch of sblock is
	component SRL16E
	generic (
     	INIT : bit_vector);
	port (
	     D   : in STD_logic;
        CE  : in STD_logic;
        CLK : in STD_logic;
        A0  : in STD_logic;
        A1  : in STD_logic;
        A2  : in STD_logic;
        A3  : in STD_logic;
        Q   : out STD_logic); 
	end component;

	component devunit
	port(
		north_type 		: in std_logic_vector(4 downto 0);
		south_type 		: in std_logic_vector(4 downto 0);
		east_type 		: in std_logic_vector(4 downto 0);
		west_type 		: in std_logic_vector(4 downto 0);
		center_type 	: in std_logic_vector(4 downto 0);
		north_state 	: in std_logic;
		south_state 	: in std_logic;
		east_state 		: in std_logic;
		west_state 		: in std_logic;
		center_state 	: in std_logic;
		rule 				: in std_logic_vector(RULE_SIZE-1 downto 0);
		rst 				: in std_logic;
		clk 				: in std_logic;          
		hit 				: out std_logic;
		result_type 	: out std_logic_vector(4 downto 0);
		result_state 	: out std_logic
		);
	end component;

--	component configcounter
--	port(
--		ce 				: in std_logic;
--		clk 				: in std_logic;
--		sset 				: in std_logic;
--		q 					: out std_logic_vector(4 downto 0)
--		);
--	end component;

	component type2lutconfig
	port (
		A					: in std_logic_VECTOR(4 downto 0);
		CLK				: in std_logic;
		D					: in std_logic_VECTOR(31 downto 0);
		WE					: in std_logic;
		SPO				: out std_logic_VECTOR(31 downto 0)
		);
	end component;

   signal lut1value 				: std_logic;							--Verdien fra den ene LUT
   signal lut2value 				: std_logic;							--Verdi av den andre LUT
   signal ff 						: std_logic;							--Flip-flop som tar vare p Sblockens tilstand
	
	signal rulehit 				: std_logic;							--Verdi fra dev-unit om regel traff eller ikke
	signal devvalue 				: std_logic;							--Verdien fra dev-unit'en, ny tilstand hvis regel har blitt aktivert	
	signal devtype 				: std_logic_vector(4 downto 0);	--Ny type fra dev-unit
	signal owntype 				: std_logic_vector(4 downto 0);	--Sblockens type, tar vare p informasjon fr denne legges ut "offentlig"

	signal config_data 			: std_logic;							--Konfigurasjonsdata for LUT
	signal config_lut1_enable 	: std_logic;							--Skriv til lut1
	signal config_lut2_enable 	: std_logic;							--Skriv til lut2
	signal luta0 					: std_logic;							--LUT-inngang 0
	signal luta1 					: std_logic;							--LUT-inngang 1
	signal luta2 					: std_logic;							--LUT-inngang 2
	signal luta3 					: std_logic;							--LUT-inngang 3
	signal newlutconfig 			: std_logic_vector(31 downto 0);	--Ny LUT-konfigurasjon
	signal count 					: std_logic_vector(4 downto 0);	--Teller som passer p hvilket konfigurasjonsbit som skrives til LUT
	signal clearcount 			: std_logic;							--Nullstill LUT-konfigurasjonstelleren

	signal rule_used_dev 		: std_logic;							--Mellomlagring av om regel har sltt til, for  unng flere regler aktivert pr dev-step

	signal t2lc_address			: std_logic_vector(4 downto 0);	--Adressen til Type-LUT-minnet, stort sett samme som owntype, bortsett nr det skrives til

begin
  	--LUT1, brukes nr Sblockens egen verdi=0
	lut1: SRL16E
		generic map (
      		INIT => X"0000")
    		port map (
      		D   => config_data,
      		CE  => config_lut1_enable,
      		A0  => luta0,
      		A1  => luta1,
      		A2  => luta2,
      		A3  => luta3,
      		Q   => lut1value,
      		CLK => clk
		);
	--LUT2, brukes nr Sblockens egen verdi=1 
	lut2: SRL16E
		generic map (
      		INIT => X"0000")
    		port map (
      		D   => config_data,
      		CE  => config_lut2_enable,
      		A0  => luta0,
      		A1  => luta1,
      		A2  => luta2,
      		A3  => luta3,
      		Q   => lut2value,
      		CLK => clk
		);
	--Regelsjekker
	dev: devunit PORT MAP(
		north_type 		=> Ntype,
		south_type 		=> Stype,
		east_type 		=> Etype,
		west_type 		=> Wtype,
		center_type 	=> owntype,
		north_state 	=> Nval,
		south_state 	=> Sval,
		east_state 		=> Eval,
		west_state 		=> Wval,
		center_state 	=> ff,
		rule 				=> rule,
		hit 				=> rulehit,
		result_type 	=> devtype,
		result_state 	=> devvalue,
		rst 				=> rst,
		clk				=> clk
	);

	--Minne for type->LUTkonfigurasjon
	t2lc : type2lutconfig port map (
		A		=> t2lc_address,
		CLK	=> clk,
		D		=> t2lc_data_in,
		WE		=> t2lc_we,
		SPO	=> newlutconfig
	);

	--Slett telleren nr Sblocken ikke konfigureres
	clearcount <= not config_enable;
  
  --Prosess for Type->LUTminnet
  process (clk,t2lc_we)
  begin
		if t2lc_we = '1' then				--Lagre konfigurasjonsdata	
			t2lc_address<=t2lc_addr_in;	--Sett addressen
		else										--Hent ut konfigdata
			t2lc_address<=owntype;			--Sett adresse=type
		end if;
  end process;

  process (rst,clk,count)
  begin
  	 if config_enable='1' then				--Konfigurere LUT: Dette setter LUT'en til  bruke 16 bit
	 		luta0 <= '1';
			luta1 <= '1';
			luta2 <= '1';
			luta3 <= '1';
	 else					 						--Sl opp i LUT, sett naboene som input
	 	 	luta0 <= Wval;
			luta1 <= Sval;
			luta2 <= Eval;
			luta3 <= Nval;
	 end if;
    if rising_edge (clk) then
      if rst = '0' then						--Reset
			owntype					<="00000";
			ff							<= '0';
			config_lut2_enable 	<= '0';
			config_lut1_enable 	<= '0';
			config_data 			<= '0';
			rule_used 				<= '0';
			rule_used_dev 			<= '0';
	 	else										--Kjr Sblock-prosesser
		 	if Step_enable = '1' then			--Step
				Outtype <= owntype;				--Typen skal synes
     	 		if ff = '0' then		 			--Egen verdi er 0, bruk LUT1
          	  	ff <= lut1value;	
          	else
	        			ff <= lut2value;			--Egen verdi er 1, bruk LUT2
     	  	   end if;
	 	  	elsif dev_enable = '1' then		--Development
				if rule_used_dev = '0' then	--Ikke brukt regel enn
					if rulehit='1' then			--Regel aktiveres
						owntype <= devtype;		--Ny type fra development
						ff <= devvalue;			--Ny state fra development
						rule_used_dev<='1';		--Har brukt regel
						rule_used<='1';			--Si fra til toplevel at denne regelen er brukt
					end if;
				else
					rule_used<='0';				--Denne regelen ble ikke brukt, skal ikke telles
				end if;
		  	elsif set_enable = '1' then		--Set-tilstanden, setter type og verdi
				owntype <= set_type;				
				ff <= set_value;
		  	elsif config_enable = '1' then	--Konfigurer LUT
				if count(4) = '0' then				--LUT1
					config_lut1_enable <= '1';
					config_lut2_enable <= '0';
					config_data <= newlutconfig(conv_integer(count));	--Hent riktig bit fra konfigurasjonen
				else								  		--LUT2
					config_lut1_enable <= '0';
					config_lut2_enable <= '1';	
					config_data <= newlutconfig(conv_integer(count));	--Hent riktig bit fra konfigurasjonen
				end if;
		  	else										--Idle-tilstanden
				config_lut1_enable <= '0';		 
				config_lut2_enable <= '0';
				rule_used <= '0';
				rule_used_dev <= '0';
				Outtype <= owntype;
			end if;
	  end if;
	end if;
	
  end process;
  Outval <= ff;									--Tilstanden synes ut
end sblock_arch;
